import java.util.*;
import java.io.*;

class Word { 
    String eng;
    String rus;
    int    count;
};

public class Dictionary { 
    static boolean LARGE_DICTIONARY = true;
    static final int FREQUENCY = LARGE_DICTIONARY ? 0 : 3;
    static final int MAX_LENGTH = LARGE_DICTIONARY ? 64 : 25;
    static final int MAX_FILE_SIZE = 16*1024;

    static final String OUTPUT = "..\\res\\";

    static int split(int[] words, int[] bestParts, int[] currParts, int offs, int partNo, int minWords) { 
        if (++partNo < currParts.length) { 
            for (int i = offs+1; i < words.length; i++) { 
                currParts[partNo-1] = i;
                minWords = split(words, bestParts, currParts, i, partNo, minWords);
            }
        } else { 
            int maxWords = 0;
            for (int i = 0, j = 0; i < currParts.length; i++) { 
                int nWords = 0;
                while (j < currParts[i]) {
                    nWords += words[j++];
                }
                if (nWords > maxWords) { 
                    maxWords = nWords;
                }
            }
            if (maxWords < minWords) { 
                System.arraycopy(currParts, 0, bestParts, 0,  currParts.length);
                minWords = maxWords;
            }            
        }
        return minWords;
    }
                
    public static void main(String[] args) throws Exception {
        String dictionaryPath = args.length > 0 ? args[0] : "Mueller7GPL.win";
	BufferedReader in = new BufferedReader(new FileReader(dictionaryPath));
	String line;
	TreeMap hash = new TreeMap();
      nextLine:
	while ((line = in.readLine()) != null) { 
	    int i = line.indexOf(' ');
	    if (i > 1) {
                for (int j = 0; j < i; j++) { 
                    char ch = line.charAt(j);
                    if (!(ch >= 'a' && ch <= 'z') && !(ch >= 'A' && ch <= 'Z')) {
                        continue nextLine;
                    }
                }                        
		Word word = new Word();                
		word.eng = line.substring(0, i).toLowerCase();
		boolean firstChar = true;
		int nBrackets = 0;
		StringBuffer buf = new StringBuffer();
		boolean skip = false;
		while (++i < line.length()) { 
		    char ch = line.charAt(i); 
		    switch (ch) { 
		    case '{':
		    case '(':
		    case '[':
			nBrackets += 1;
			continue;
		    case '}':
		    case ')':
		    case ']':
			nBrackets -= 1;
			continue;
		    case '_':
			skip = true;
			continue;
		    case ' ':
			skip = false;
			if (!firstChar && nBrackets == 0) { 
			    buf.append(' ');
			}
			continue;
		    default:
			if (!skip && nBrackets == 0) {
			    if (Character.isLetter(ch) && ch > 'z') { 
				firstChar = false;
				buf.append(ch);
			    } else if (!firstChar) { 
				word.rus = buf.toString().trim();
				int len = word.rus.length();
				if (len > 2 && len <= MAX_LENGTH ) {
                                    if (hash.get(word.eng) == null) { 
                                        hash.put(word.eng, word);
                                    }
				}
				continue nextLine;
			    }
			}
		    }
		}
	    }
	}
        if (FREQUENCY != 0) { 
            byte[] buf = new byte[1025*1024];
            int rc;
            FileInputStream fs = new FileInputStream("samples.txt");
            while ((rc = fs.read(buf)) > 0) { 
                for (int i = 0; i < rc; i++) { 
                    int j = i;
                    char ch;
                    while (j < rc && (((ch = (char)buf[j]) >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z'))) {
                        j += 1;
                    }
                    if (j > i) { 
                        String key = new String(buf, i, j-i).toLowerCase();
                        Word word = (Word)hash.get(key);
                        if (word != null) { 
                            word.count += 1;
                        }
                        i = j;
                    }
                }
            }
        }
        if (args.length > 0) { 
            int nParts = Integer.parseInt(args[0]);
            int partNo = Integer.parseInt(args[1]);
            Iterator iterator = hash.values().iterator();
            char firstLetter = '\0';
            int[] words = new int[26];
            int i = -1;
            while (iterator.hasNext()) { 
                Word word = (Word)iterator.next();
                if (word.eng.charAt(0) != firstLetter) {
                    firstLetter = word.eng.charAt(0);
                    i += 1;
                }
                words[i] += 1;
            }
            int[] currParts = new int[nParts];
            int[] bestParts = new int[nParts];
            currParts[nParts-1] = words.length;
            split(words, bestParts, currParts, 0, 0, Integer.MAX_VALUE);
            iterator = hash.values().iterator();
            i = 0;
            if (partNo > 1) { 
                for (int n = bestParts[partNo-2]; i < n; i++) { 
                    for (int j = words[i]; --j >= 0;) { 
                        iterator.next();
                    }
                }    
            }        
            TreeMap subset = new TreeMap();
            for (int n = bestParts[partNo-1]; i < n; i++) { 
                for (int j = words[i]; --j >= 0;) { 
                    Word word = (Word)iterator.next();
                    subset.put(word.eng, word);
                }
            }
            hash = subset;
        }
	FileOutputStream index = new FileOutputStream(OUTPUT + "dictionary.idx");
	FileOutputStream out = new FileOutputStream(OUTPUT + "volume.1");
	Iterator iterator = hash.values().iterator();
	int n = 0;
	int size = 0;
	int totalSize = 0;
        int id = 1;
        String prevWord = null;
        StringBuffer ib = new StringBuffer();
	while (iterator.hasNext()) { 
	    Word word = (Word)iterator.next();
	    if (word.count >= FREQUENCY) { 
		StringBuffer sb = new StringBuffer();
                String thisWord = word.eng;
                if (prevWord == null) { 
                    ib.append(thisWord);
                    ib.append(':');
                    sb.append(thisWord);
                } else { 
                    int prefix;
                    int len = thisWord.length() > prevWord.length() ? prevWord.length() : thisWord.length();
                    if (len > 9) { 
                        len = 9;
                    }
                    for (prefix = 0; prefix < len && thisWord.charAt(prefix) == prevWord.charAt(prefix); prefix++);
                    if (prefix > 0) { 
                        sb.append((char)(prefix + '0'));
                        sb.append(thisWord.substring(prefix));
                    } else { 
                        sb.append(thisWord);
                    }
                }
		sb.append(':');
		sb.append(word.rus);
		//sb.append('\r');
		sb.append('\n');
		byte[] bytes = sb.toString().getBytes();
                if (size + bytes.length > MAX_FILE_SIZE) { 
                    ib.append(prevWord);
                    ib.append('\n');
                    index.write(ib.toString().getBytes());
                    ib = new StringBuffer();
                    ib.append(thisWord);
                    ib.append(':');
                    out.close();
                    out = new FileOutputStream(OUTPUT + "volume." + ++id);
                    size = 0;
                }
		size += bytes.length;
                totalSize += bytes.length;
		n += 1;
                prevWord = thisWord;
		out.write(bytes);
	    }
	}
	out.close();
        ib.append(prevWord);
        ib.append('\n');
        index.write(ib.toString().getBytes());
        index.close();
	System.out.println("Words " + n + ", size " + totalSize);
    }
}